"""
  Generated by Eclipse Cyclone DDS idlc Python Backend
  Cyclone DDS IDL version: v0.9.0
  Module: DDS.XTypes

"""

from ._ddsi_xt_type_object import ANNOTATION_OCTETSEC_VALUE_MAX_LEN, ANNOTATION_STR_VALUE_MAX_LEN, AliasMemberFlag, AliasTypeFlag, AnnotationParameterFlag, AnnotationParameterValue, AnnotationTypeFlag, AppliedAnnotation, AppliedAnnotationParameter, AppliedAnnotationParameterSeq, AppliedAnnotationSeq, AppliedBuiltinMemberAnnotations, AppliedBuiltinTypeAnnotations, AppliedVerbatimAnnotation, BitBound, BitflagFlag, BitmaskTypeFlag, BitsetMemberFlag, BitsetTypeFlag, CollectionElementFlag, CollectionTypeFlag, CommonAliasBody, CommonAnnotationParameter, CommonArrayHeader, CommonBitfield, CommonBitflag, CommonBitmaskHeader, CommonCollectionElement, CommonCollectionHeader, CommonDiscriminatorMember, CommonEnumeratedHeader, CommonEnumeratedLiteral, CommonStructMember, CommonUnionMember, CompleteAliasBody, CompleteAliasHeader, CompleteAliasType, CompleteAnnotationHeader, CompleteAnnotationParameter, CompleteAnnotationParameterSeq, CompleteAnnotationType, CompleteArrayHeader, CompleteArrayType, CompleteBitfield, CompleteBitfieldSeq, CompleteBitflag, CompleteBitflagSeq, CompleteBitmaskHeader, CompleteBitmaskType, CompleteBitsetHeader, CompleteBitsetType, CompleteCollectionElement, CompleteCollectionHeader, CompleteDiscriminatorMember, CompleteElementDetail, CompleteEnumeratedHeader, CompleteEnumeratedLiteral, CompleteEnumeratedLiteralSeq, CompleteEnumeratedType, CompleteExtendedType, CompleteMapType, CompleteMemberDetail, CompleteSequenceType, CompleteStructHeader, CompleteStructMember, CompleteStructMemberSeq, CompleteStructType, CompleteTypeDetail, CompleteTypeObject, CompleteUnionHeader, CompleteUnionMember, CompleteUnionMemberSeq, CompleteUnionType, EK_BOTH, EK_COMPLETE, EK_MINIMAL, EnumTypeFlag, EnumeratedLiteralFlag, EquivalenceHash, EquivalenceKind, ExtendedAnnotationParameterValue, ExtendedTypeDefn, INVALID_LBOUND, INVALID_SBOUND, LBound, LBoundSeq, MEMBER_NAME_MAX_LENGTH, MemberFlag, MemberFlagMinimalMask, MemberId, MemberName, MinimalAliasBody, MinimalAliasHeader, MinimalAliasType, MinimalAnnotationHeader, MinimalAnnotationParameter, MinimalAnnotationParameterSeq, MinimalAnnotationType, MinimalArrayHeader, MinimalArrayType, MinimalBitfield, MinimalBitfieldSeq, MinimalBitflag, MinimalBitflagSeq, MinimalBitmaskHeader, MinimalBitmaskType, MinimalBitsetHeader, MinimalBitsetType, MinimalCollectionElement, MinimalCollectionHeader, MinimalDiscriminatorMember, MinimalEnumeratedHeader, MinimalEnumeratedLiteral, MinimalEnumeratedLiteralSeq, MinimalEnumeratedType, MinimalExtendedType, MinimalMapType, MinimalMemberDetail, MinimalSequenceType, MinimalStructHeader, MinimalStructMember, MinimalStructMemberSeq, MinimalStructType, MinimalTypeDetail, MinimalTypeObject, MinimalUnionHeader, MinimalUnionMember, MinimalUnionMemberSeq, MinimalUnionType, NameHash, PlainArrayLElemDefn, PlainArraySElemDefn, PlainCollectionHeader, PlainMapLTypeDefn, PlainMapSTypeDefn, PlainSequenceLElemDefn, PlainSequenceSElemDefn, PrimitiveTypeId, QualifiedTypeName, SBound, SBoundSeq, StringLTypeDefn, StringSTypeDefn, StronglyConnectedComponent, StronglyConnectedComponentId, StructMemberFlag, StructTypeFlag, TI_PLAIN_ARRAY_LARGE, TI_PLAIN_ARRAY_SMALL, TI_PLAIN_MAP_LARGE, TI_PLAIN_MAP_SMALL, TI_PLAIN_SEQUENCE_LARGE, TI_PLAIN_SEQUENCE_SMALL, TI_STRING16_LARGE, TI_STRING16_SMALL, TI_STRING8_LARGE, TI_STRING8_SMALL, TI_STRONGLY_CONNECTED_COMPONENT, TK_ALIAS, TK_ANNOTATION, TK_ARRAY, TK_BITMASK, TK_BITSET, TK_BOOLEAN, TK_BYTE, TK_CHAR16, TK_CHAR8, TK_ENUM, TK_FLOAT128, TK_FLOAT32, TK_FLOAT64, TK_INT16, TK_INT32, TK_INT64, TK_MAP, TK_NONE, TK_SEQUENCE, TK_STRING16, TK_STRING8, TK_STRUCTURE, TK_UINT16, TK_UINT32, TK_UINT64, TK_UNION, TYPE_NAME_MAX_LENGTH, TypeFlag, TypeFlagMinimalMask, TypeIdentfierWithSizeSeq, TypeIdentiferKind, TypeIdentifier, TypeIdentifierPair, TypeIdentifierPairSeq, TypeIdentifierSeq, TypeIdentifierTypeObjectPair, TypeIdentifierTypeObjectPairSeq, TypeIdentifierWithDependencies, TypeIdentifierWithDependenciesSeq, TypeIdentifierWithSize, TypeInformation, TypeInformationSeq, TypeKind, TypeObject, TypeObjectHashId, TypeObjectSeq, UnionCaseLabelSeq, UnionDiscriminatorFlag, UnionMemberFlag, UnionTypeFlag
from ._ddsi_xt_typemap import TypeMapping
__all__ = ["ANNOTATION_OCTETSEC_VALUE_MAX_LEN", "ANNOTATION_STR_VALUE_MAX_LEN", "AliasMemberFlag", "AliasTypeFlag", "AnnotationParameterFlag", "AnnotationParameterValue", "AnnotationTypeFlag", "AppliedAnnotation", "AppliedAnnotationParameter", "AppliedAnnotationParameterSeq", "AppliedAnnotationSeq", "AppliedBuiltinMemberAnnotations", "AppliedBuiltinTypeAnnotations", "AppliedVerbatimAnnotation", "BitBound", "BitflagFlag", "BitmaskTypeFlag", "BitsetMemberFlag", "BitsetTypeFlag", "CollectionElementFlag", "CollectionTypeFlag", "CommonAliasBody", "CommonAnnotationParameter", "CommonArrayHeader", "CommonBitfield", "CommonBitflag", "CommonBitmaskHeader", "CommonCollectionElement", "CommonCollectionHeader", "CommonDiscriminatorMember", "CommonEnumeratedHeader", "CommonEnumeratedLiteral", "CommonStructMember", "CommonUnionMember", "CompleteAliasBody", "CompleteAliasHeader", "CompleteAliasType", "CompleteAnnotationHeader", "CompleteAnnotationParameter", "CompleteAnnotationParameterSeq", "CompleteAnnotationType", "CompleteArrayHeader", "CompleteArrayType", "CompleteBitfield", "CompleteBitfieldSeq", "CompleteBitflag", "CompleteBitflagSeq", "CompleteBitmaskHeader", "CompleteBitmaskType", "CompleteBitsetHeader", "CompleteBitsetType", "CompleteCollectionElement", "CompleteCollectionHeader", "CompleteDiscriminatorMember", "CompleteElementDetail", "CompleteEnumeratedHeader", "CompleteEnumeratedLiteral", "CompleteEnumeratedLiteralSeq", "CompleteEnumeratedType", "CompleteExtendedType", "CompleteMapType", "CompleteMemberDetail", "CompleteSequenceType", "CompleteStructHeader", "CompleteStructMember", "CompleteStructMemberSeq", "CompleteStructType", "CompleteTypeDetail", "CompleteTypeObject", "CompleteUnionHeader", "CompleteUnionMember", "CompleteUnionMemberSeq", "CompleteUnionType", "EK_BOTH", "EK_COMPLETE", "EK_MINIMAL", "EnumTypeFlag", "EnumeratedLiteralFlag", "EquivalenceHash", "EquivalenceKind", "ExtendedAnnotationParameterValue", "ExtendedTypeDefn", "INVALID_LBOUND", "INVALID_SBOUND", "LBound", "LBoundSeq", "MEMBER_NAME_MAX_LENGTH", "MemberFlag", "MemberFlagMinimalMask", "MemberId", "MemberName", "MinimalAliasBody", "MinimalAliasHeader", "MinimalAliasType", "MinimalAnnotationHeader", "MinimalAnnotationParameter", "MinimalAnnotationParameterSeq", "MinimalAnnotationType", "MinimalArrayHeader", "MinimalArrayType", "MinimalBitfield", "MinimalBitfieldSeq", "MinimalBitflag", "MinimalBitflagSeq", "MinimalBitmaskHeader", "MinimalBitmaskType", "MinimalBitsetHeader", "MinimalBitsetType", "MinimalCollectionElement", "MinimalCollectionHeader", "MinimalDiscriminatorMember", "MinimalEnumeratedHeader", "MinimalEnumeratedLiteral", "MinimalEnumeratedLiteralSeq", "MinimalEnumeratedType", "MinimalExtendedType", "MinimalMapType", "MinimalMemberDetail", "MinimalSequenceType", "MinimalStructHeader", "MinimalStructMember", "MinimalStructMemberSeq", "MinimalStructType", "MinimalTypeDetail", "MinimalTypeObject", "MinimalUnionHeader", "MinimalUnionMember", "MinimalUnionMemberSeq", "MinimalUnionType", "NameHash", "PlainArrayLElemDefn", "PlainArraySElemDefn", "PlainCollectionHeader", "PlainMapLTypeDefn", "PlainMapSTypeDefn", "PlainSequenceLElemDefn", "PlainSequenceSElemDefn", "PrimitiveTypeId", "QualifiedTypeName", "SBound", "SBoundSeq", "StringLTypeDefn", "StringSTypeDefn", "StronglyConnectedComponent", "StronglyConnectedComponentId", "StructMemberFlag", "StructTypeFlag", "TI_PLAIN_ARRAY_LARGE", "TI_PLAIN_ARRAY_SMALL", "TI_PLAIN_MAP_LARGE", "TI_PLAIN_MAP_SMALL", "TI_PLAIN_SEQUENCE_LARGE", "TI_PLAIN_SEQUENCE_SMALL", "TI_STRING16_LARGE", "TI_STRING16_SMALL", "TI_STRING8_LARGE", "TI_STRING8_SMALL", "TI_STRONGLY_CONNECTED_COMPONENT", "TK_ALIAS", "TK_ANNOTATION", "TK_ARRAY", "TK_BITMASK", "TK_BITSET", "TK_BOOLEAN", "TK_BYTE", "TK_CHAR16", "TK_CHAR8", "TK_ENUM", "TK_FLOAT128", "TK_FLOAT32", "TK_FLOAT64", "TK_INT16", "TK_INT32", "TK_INT64", "TK_MAP", "TK_NONE", "TK_SEQUENCE", "TK_STRING16", "TK_STRING8", "TK_STRUCTURE", "TK_UINT16", "TK_UINT32", "TK_UINT64", "TK_UNION", "TYPE_NAME_MAX_LENGTH", "TypeFlag", "TypeFlagMinimalMask", "TypeIdentfierWithSizeSeq", "TypeIdentiferKind", "TypeIdentifier", "TypeIdentifierPair", "TypeIdentifierPairSeq", "TypeIdentifierSeq", "TypeIdentifierTypeObjectPair", "TypeIdentifierTypeObjectPairSeq", "TypeIdentifierWithDependencies", "TypeIdentifierWithDependenciesSeq", "TypeIdentifierWithSize", "TypeInformation", "TypeInformationSeq", "TypeKind", "TypeObject", "TypeObjectHashId", "TypeObjectSeq", "UnionCaseLabelSeq", "UnionDiscriminatorFlag", "UnionMemberFlag", "UnionTypeFlag", "TypeMapping", ]
